Comet.Popovers = function() {

    "use strict";

    var el;
    var $el;
    var isTooltip;
    var elTitle;
    var debouncedEls = [];  // debounce variable case of click then focus
    var nubSideLength = 12;
    var nubSpacing = 0; // space between nub and target element in px
    var nubEdgeOffset = 6; //distace offset from the edge of the tooltip
    var nubBaseLength = nubSideLength * Math.sqrt(2);
    var nubHeight = nubBaseLength / 2;
    var defaultTooltipPlacement = "top"; // default to comet-tooltip--top if no placement is specified
    var defaultPopoverPlacement = "bottom"; // default to comet-popover--bottom if no placement is specified

    var myHorizontal;  // horizontal alignment of tooltip with target element
    var myVertical;  // vertical alignment of tooltip with target element
    var atHorizontal; // horizontal position on target element tooltip is aligned with
    var atVertical; // vertical position on target element tooltip is aligned with

    var directions = ["top", "right", "bottom", "left"];

    var initialize = function(event, initializeTooltips) {
        // TODO: fix multiple initializations
        deactivateExistingPopovers(initializeTooltips);

        if (initializeTooltips) {
            // initialize tooltips only
            $('.js-comet-tooltip-trigger, *[data-comet-tooltip-text]').on("mouseenter mouseleave", triggerPopoverEvent);

            // initialize tooltips to display on load
            $('.js-comet-tooltip-trigger.js-comet-tooltip-trigger--on-load, *[data-comet-tooltip-text].js-comet-tooltip-trigger--on-load').trigger("mouseenter");
        } else {
            // initialize general popovers only
            $('.js-comet-popover-trigger').each(function() {
                var isHover = $(this).data("comet-popover-event") === "hover";
                var events = isHover ? "mouseenter mouseleave" : "click focus";
                $(this).on(events, triggerPopoverEvent);
            });

            $('.comet-popover').on("show hide", handlePopoverTrigger);
            $('.js-comet-tooltip-trigger, *[data-comet-tooltip-text]').on("show hide", ".comet-tooltip__content", handlePopoverTrigger);
        }

        listenForClickAway();
        listenForEscKeyPress();
    };

    var deactivateExistingPopovers = function(initializeTooltips) {
        if (initializeTooltips) {
            $('.js-comet-tooltip-trigger, *[data-comet-tooltip-text]').off("mouseenter mouseleave"); //remove listeners if already initialized
            $('.comet-tooltip__content').remove();
        } else {
            $('.js-comet-popover-trigger').off();
        }
    };

    var triggerPopoverEvent = function(event) {
        event.stopPropagation();
        var $popover;

        isTooltip = !!$(this).data("comet-tooltip-text");
        var debounce = event.type !== "mouseenter";

        var eventOpts = {
            "el": event.currentTarget,
            "debounce": debounce
        };

        if (isTooltip && (event.type === "mouseenter" || event.type === "mouseleave")) {
            positionTargetElement(this);

            if (!tooltipHasBeenRendered(this)) {
                renderTooltip(this);
            }

            $popover = $(this).find('.comet-tooltip__content');
        } else {
            var popoverId = $(this).data("comet-popover");
            $popover = $('#' + popoverId);
            // hide any tooltips
            $(this).find('.comet-tooltip__content').trigger("hide", eventOpts);
        }

        if (popoverIsVisible($popover)) {
            if (isTooltip && event.type === "mouseenter") {
                return;  //tooltips should never hide on mouseenter
            }
            $popover.trigger("hide", eventOpts);
        } else {
            if (isTooltip && event.type === "mouseleave") {
                return;  //tooltips should never show on mouseleave
            }
            $popover.trigger("show", eventOpts);
        }
    };

    var handlePopoverTrigger = function(event, opts) {
        opts = opts || {};
        el = opts.el || el;
        $el = $(el);
        var debounce = opts.debounce === void 0 ? true : opts.debounce;

        setTimeout(function() {
            debouncedEls = [];
        }, 200);

        if (debouncedEls.indexOf(this) === -1) {
            // prevent this element from accepting show and hide events for 200ms
            // this allows both focus and click events to fire and the show event
            // the tooltip is shown but not hidden
            if (debounce) {
                debouncedEls.push(this);
            }

            var $popover = $(this);

            if (event.type === "hide") {
                hidePopover($popover);
            } else if (event.type === "show") {
                showPopover($popover);
            }
        }
    };

    var popoverIsVisible = function($popover) {
        return $popover.hasClass("visible");
    };

    var listenForClickAway = function() {
        $(document).on("click", function(event) {
            if (!isPopover(event.target)) {
                closeAllPopups();
            }
        });
    };

    var listenForEscKeyPress = function() {
        $(document).on("keyup", function(event) {
            var escKeyCode = 27;
            if (event.which === escKeyCode) {
                closeAllPopups();
            }
        });
    };

    var closeAllPopups = function() {
        $('.comet-popover').trigger("hide");
    };

    var isPopover = function(clickedEl) {
        return isOrHasParent(clickedEl, "comet-popover");
    };

    var isOrHasParent = function(clickedEl, className) {
        var selector = "." + className;
        return $(clickedEl).hasClass(className) || !!$(clickedEl).closest(selector).length;
    };

    var showPopover = function($popover) {
        if (!isTooltip) {
            // hide any showing popovers
            $('.comet-popover, .comet-tooltip__content').not($popover).trigger("hide");
        }

        removeTitleAttr(); // prevent native browser tooltip

        el.offsetHeight; //Forces a redraw so the tooltip is added in its invisible state before the "visible" class is added
        $popover.addClass("visible");

        var positionInfo = getPositionInfo($popover);
        positionPopover($popover, positionInfo);
    };

    var removeTitleAttr = function() {
        elTitle = "";
        elTitle = $el.attr("title");

        if (elTitle) {
            $el.attr("title", "");
        }
    };

    var hidePopover = function($popover) {
        $popover.removeClass("visible");
        addTitleAttr(); // add title back for accessibility
    };

    var addTitleAttr = function() {
        if (elTitle) {
            $el.attr("title", elTitle);
        }
    };

    var positionTargetElement = function(targetEl) {
        if ($(targetEl).css("position") === "static") {  // only position if element is not already positioned
            $(targetEl).css("position", "relative");
        }
    };

    var tooltipHasBeenRendered = function(targetEl) {
        return !!$(targetEl).find('.comet-tooltip__content').length;
    };

    var renderTooltip = function(targetEl) {

        $(targetEl).append('<div class="comet-tooltip__content"><span class="comet-tooltip__nub"></span><span class="comet-tooltip__text">' + $(targetEl).data("comet-tooltip-text") + '</span></div>');
    };

    var getPositionInfo = function($popover) {

        var placement = getPlacement($popover);

        if (!placement) {

            var defaultClass = isTooltip ? "comet-tooltip--" + defaultTooltipPlacement : "comet-popover--" + defaultPopoverPlacement;
            $(getPlacedEl($popover)).addClass(defaultClass);
            placement = isTooltip ? defaultTooltipPlacement : defaultPopoverPlacement;
        }

        var tooltipOffset = nubHeight + nubSpacing; //shift to account for nub and spacing

        if (isVertical(placement)) {
            // the tooltip is above or below the target element
            myHorizontal = "center";
            atHorizontal = "center";
            myVertical = flipDirection(placement) + buildOffsetString(placement, tooltipOffset);
            atVertical = placement;
        } else {
            // tooltip is to the left or right of target element
            myVertical = "center";
            atVertical = "center";
            myHorizontal = flipDirection(placement) + buildOffsetString(placement, tooltipOffset);
            atHorizontal = placement;
        }

        var tooltipShiftDirection = getTooltipShiftDirection($popover);

        if (tooltipShiftDirection) {
            var myAlignment = flipDirection(tooltipShiftDirection);

            var nubPlacementShift = nubEdgeOffset + nubBaseLength / 2;
            myAlignment = myAlignment + buildOffsetString(myAlignment, nubPlacementShift);

            if (isVertical(tooltipShiftDirection)) {
                myVertical = myAlignment;
            } else {
                myHorizontal = myAlignment;
            }
        }

        // position using jQuery UI's position() function
        return {
            "my": myHorizontal + " " + myVertical,
            "at": atHorizontal + " " + atVertical,
            "of": $el,
            "using": setFinalPosition
        };
    };

    var getPlacement = function($popover) {
        var placement;

        directions.forEach(function(direction) {
            var pattern = new RegExp("comet-.*--" + direction);
            if (pattern.test(getPlacedEl($popover).className)) {
                placement = direction;
            }
        });

        return placement || null; // if no placement is specified, default to top
    };

    var getPlacedEl = function($popover) {
        // return the element that contains the positioning class
        return isTooltip ? el : $popover.get(0);
    };

    var isVertical = function(direction) {
        return direction === "top" || direction === "bottom";
    };

    var flipDirection = function(direction) {
        var directionOpposites = {
            "top": "bottom",
            "bottom": "top",
            "left": "right",
            "right": "left"
        };

        return directionOpposites[direction];
    };

    var buildOffsetString = function(direction, tooltipOffset) {
        //create string to set distance between tooltip and target element to  i.e. "top-20"

        var operator;

        if (direction === "top" || direction === "left") {
            operator = "-";
        } else {
            operator = "+";
        }

        return operator + tooltipOffset;
    };

    var getTooltipShiftDirection = function($popover) {
        var shiftDirection;

        directions.forEach(function(direction) {
            var pattern = new RegExp("comet-(tooltip|popover)--.+-" + direction);
            if (pattern.test(getPlacedEl($popover).className)) {
                shiftDirection = direction;
            }
        });

        return shiftDirection || null;
    };

    var positionPopover = function($popover, positionInfo) {
        // HACK: must call position twice for <span> and <a> elements
        $popover.position(positionInfo).position(positionInfo);
    };

    var setFinalPosition = function(position, info) {
        $(this).find(".comet-popover__nub , .comet-tooltip__nub").toggleClass("flipped", isFlipped(info));

        // have to set final position manually because of position's using() callback
        $(this).css(position);
    };

    var isFlipped = function(info) {
        // find placement of tooltip relative to target element and see if it matches intended placement
        var tooltipVerticalPlacement = flipDirection(info.vertical);
        var tooltipHorizontalPlacement = flipDirection(info.horizontal);
        return myVertical.indexOf(tooltipVerticalPlacement) > -1 || myHorizontal.indexOf(tooltipHorizontalPlacement) > -1;
    };

    // initialize tooltips by default when DOM content has loaded
    $(document).ready(initialize);

    return {
        "initialize": initialize
    };
}();

